@@ -15,18 +15,24 @@
 # Author: James Diprose, Aniek Roelofs, Tuan Chien
 
 import os
+import subprocess
+import sys
 
 import click
-
 from observatory.platform.cli.click_utils import INDENT1, INDENT2, INDENT3, indent
 from observatory.platform.cli.generate_command import GenerateCommand
 from observatory.platform.cli.platform_command import PlatformCommand
 from observatory.platform.cli.terraform_command import TerraformCommand
-from observatory.platform.platform_builder import DEBUG, HOST_GID, HOST_UID
+from observatory.platform.observatory_config import (
+    generate_fernet_key,
+    generate_secret_key,
+)
+from observatory.platform.platform_builder import DEBUG, HOST_UID
+from observatory.platform.utils.config_utils import observatory_home
 from observatory.platform.utils.config_utils import (
     terraform_credentials_path as default_terraform_credentials_path,
-    observatory_home,
 )
+from observatory.platform.utils.proc_utils import stream_process
 
 PLATFORM_NAME = "Observatory Platform"
 TERRAFORM_NAME = "Observatory Terraform"
@@ -64,21 +70,8 @@ def cli():
     help="The user id of the host system. Used to set the user id in the Docker containers.",
     show_default=True,
 )
-@click.option(
-    "--host-gid",
-    type=click.INT,
-    default=HOST_GID,
-    help="The group id of the host system. Used to set the group id in the Docker containers.",
-    show_default=True,
-)
 @click.option("--debug", is_flag=True, default=DEBUG, help="Print debugging information.")
-def platform(
-    command: str,
-    config_path: str,
-    host_uid: int,
-    host_gid: int,
-    debug,
-):
+def platform(command: str, config_path: str, host_uid: int, debug):
     """Run the local Observatory Platform platform.\n
 
     COMMAND: the command to give the platform:\n
@@ -90,12 +83,7 @@ def platform(
     print(f"{PLATFORM_NAME}: checking dependencies...".ljust(min_line_chars), end="\r")
     if os.path.isfile(config_path):
         # Make the platform command, which encapsulates functionality for running the observatory
-        platform_cmd = PlatformCommand(
-            config_path,
-            host_uid=host_uid,
-            host_gid=host_gid,
-            debug=debug,
-        )
+        platform_cmd = PlatformCommand(config_path, host_uid=host_uid, debug=debug)
 
         # Check dependencies
         platform_check_dependencies(platform_cmd, min_line_chars=min_line_chars)
@@ -111,7 +99,7 @@ def platform(
         print(indent("config.yaml:", INDENT1))
         print(indent(f"- file not found, generating a default file on path: {config_path}", INDENT2))
         generate_cmd = GenerateCommand()
-        generate_cmd.generate_local_config(config_path)
+        generate_cmd.generate_local_config(config_path, editable=False, workflows=[], oapi=False)
         exit(os.EX_CONFIG)
 
 
@@ -226,73 +214,184 @@ def generate():
     COMMAND: the commands to run include:\n
       - secrets: generate secrets.\n
       - config: generate configuration files for the Observatory Platform.\n
+      - project: generate a new project directory and required files.\n
+      - workflow: generate all files for a new workflow.
     """
 
     pass
 
 
 @generate.command()
-@click.argument(
-    "project_path",
-    type=click.Path(exists=True, file_okay=False, dir_okay=True),
-)
+@click.argument("project_path", type=click.Path(exists=False, file_okay=False, dir_okay=True))
 @click.argument("package_name", type=str)
-@click.argument("workflow_type", type=click.Choice(["Workflow", "StreamTelescope", "SnapshotTelescope"]))
-@click.argument("workflow_name", type=str)
-def workflow(project_path: str, package_name: str, workflow_type: str, workflow_name: str):
-    """Generate a new workflow.
+@click.argument("author_name", type=str)
+def project(project_path: str, package_name: str, author_name: str):
+    """Generate a new workflows project.
 
     \b
     PROJECT_PATH is the Python project path.
     PACKAGE_NAME is the Python package name.
-    WORKFLOW_TYPE is the type of workflow.
-    WORKFLOW_NAME is the the name of your new workflow.
+    AUTHOR_NAME is the author name, used for readthedocs.
 
     \b
-    For example, the command: observatory generate workflow /path/to/my-workflows-project my_workflows_project Workflow MyWorkflow
+    For example, the command: observatory generate project /path/to/my-workflows-project my_workflows_project
 
     \b Will generate the following files and folders:
 
     \b
     └── my-workflows-project
         ├── docs
+        │   ├── _build
+        │   ├── _static
+        │   ├── _templates
+        │   ├── workflows
+        │   ├── conf.py
+        │   ├── generate_schema_csv.py
         │   ├── index.rst
-        │   └── my_workflow.md
+        │   ├── make.bat
+        │   ├── Makefile
+        │   ├── requirements.txt
+        │   └── test_generate_schema_csv.py
+        ├── {package_name}
+        │   ├── dags
+        │   │   └── __init__.py
+        │   ├── database
+        │   │   ├── schema
+        │   │   │   └── __init__.py
+        │   │   └── __init__.py
+        │   └── workflows
+        │   │   ├── tests
+        │   │   │   └── __init__.py
+        │   │   └── __init__.py
+        │   ├── __init__.py
+        │   └── config.py
+        ├── setup.cfg
+        └── setup.py
+    """
+    cmd = GenerateCommand()
+    cmd.generate_workflows_project(project_path, package_name, author_name)
+
+    if click.confirm(
+        f"Would you like to install the '{package_name}' package inside your new project? This is required for "
+        f"the workflows to function."
+    ):
+        proc = subprocess.Popen(
+            [sys.executable, "-m", "pip", "install", "-e", "."],
+            stdout=subprocess.PIPE,
+            stderr=subprocess.PIPE,
+            cwd=project_path,
+        )
+        stream_process(proc, True)
+
+
+@generate.command()
+@click.argument(
+    "workflow_type", type=click.Choice(["Workflow", "StreamTelescope", "SnapshotTelescope", "OrganisationTelescope"])
+)
+@click.argument("workflow_name", type=str)
+@click.option(
+    "-p",
+    "--project-path",
+    type=click.Path(exists=True, file_okay=False, dir_okay=True),
+    default=os.getcwd(),
+    help="The path to the project directory.",
+    show_default=True,
+)
+def workflow(workflow_type: str, workflow_name: str, project_path: str):
+    """Generate all files for a new workflow.
+
+    \b
+    WORKFLOW_TYPE is the type of workflow.
+    WORKFLOW_NAME is the the name of your new workflow.
+    PROJECT_PATH is the Python project path, default is current directory.
+
+    \b
+    For example, the command: observatory generate workflow /path/to/my-workflows-project my_workflows_project Workflow MyWorkflow
+
+    \b Will generate the files in bold inside your workflows project:
+
+    \b
+    └── my-workflows-project
+        ├── docs
+        │   ├── _build
+        │   ├── _static
+        │   ├── _templates
+        │   ├── workflows
+        │   │   └── \033[1mmy_workflow.md\033[0m
+        │   ├── generate_schema_csv.py
+        │   ├── index.rst
+        │   ├── make.bat
+        │   └── Makefile
         ├── my_workflows_project
         │   ├── dags
         │   │   ├── __init__.py
-        │   │   └── my_workflow.py
-        │   ├── __init__.py
+        │   │   └── \033[1mmy_workflow.py\033[0m
+        │   ├── database
+        │   │   ├── schema
+        │   │   │   ├── __init__.py
+        │   │   │   └── \033[1mmy_workflow_YYYY_MM_DD.json\033[0m
+        │   │   └── __init__.py
         │   └── workflows
-        │       ├── __init__.py
-        │       └── my_workflow.py
-        └── tests
-            └── my_workflows_project
-                ├── __init__.py
-                └── workflows
-                    ├── __init__.py
-                    └── test_my_workflow.py
+        │   │   ├── tests
+        │   │   │   ├── __init__.py
+        │   │   │   └── \033[1mtest_my_workflow.py\033[0m
+        │   │   ├── __init__.py
+        │   │   └── \033[1mmy_workflow.py\033[0m
+        │   ├── __init__.py
+        │   ├── config.py
+        │   └── \033[1midentifiers.py\033[0m (OrganisationTelescope only)
+        ├── setup.cfg
+        └── setup.py
     """
+    print(f"Given workflow type: {workflow_type}")
+    print(f"Given workflow name: {workflow_name}")
+    print(f"Given project path: {project_path}\n")
+
+    # Check if egg-info dir is available
+    egg_info_dir = [d for d in os.listdir(project_path) if d.endswith(".egg-info")]
+    if not egg_info_dir:
+        print(
+            "Invalid workflows project, the given projects directory does not contain an installed python package.\n"
+            "Either run this command from inside a valid workflows project or specify the path to a valid workflows "
+            "project with the '--project-path' option.\n"
+            "A new workflows project can be created using the 'observatory generate project' command."
+        )
+        exit(os.EX_CONFIG)
+    if len(egg_info_dir) > 1:
+        print(
+            "Invalid workflows project, the given projects directory contains more than 1 installed python package.\n"
+            "Either run this command from inside a valid workflows project or specify the path to a valid workflows "
+            "project with the '--project-path' option.\n"
+            "A new workflows project can be created using the 'observatory generate project' command."
+        )
+        exit(os.EX_CONFIG)
+
+    # Get package name
+    top_level_file = os.path.join(project_path, egg_info_dir[0], "top_level.txt")
+    with open(top_level_file, "r") as f:
+        package_name = f.readline().strip()
+        print(f"Found package inside project: {package_name}")
 
-    # TODO: project path and package name should be inferred somehow
     cmd = GenerateCommand()
-    cmd.generate_new_workflow(
-        project_path=project_path, package_name=package_name, workflow_type=workflow_type, workflow_name=workflow_name
+    cmd.generate_workflow(
+        project_path=project_path, package_name=package_name, workflow_type=workflow_type, workflow_class=workflow_name
     )
 
 
 @generate.command()
-@click.argument("command", type=click.Choice(["fernet-key"]))
+@click.argument("command", type=click.Choice(["fernet-key", "secret-key"]))
 def secrets(command: str):
     """Generate secrets for the Observatory Platform.\n
 
     COMMAND: the type of secret to generate:\n
       - fernet-key: generate a random Fernet Key.\n
+      - secret-key: generate a random secret key.\n
     """
 
     if command == "fernet-key":
-        cmd = GenerateCommand()
-        print(cmd.generate_fernet_key())
+        print(generate_fernet_key())
+    else:
+        print(generate_secret_key())
 
 
 @generate.command()
@@ -304,34 +403,63 @@ def secrets(command: str):
     help="The path to the config file to generate.",
     show_default=True,
 )
-def config(command: str, config_path: str):
+@click.option("--interactive", flag_value=True, help="Configuration through an interactive Q&A mode")
+@click.option(
+    "--ao-wf",
+    flag_value=True,
+    help="Indicates that the academic-observatory-workflows was installed through the cli installer script",
+)
+@click.option(
+    "--oaebu-wf",
+    flag_value=True,
+    help="Indicates that the oaebu-workflows was installed through the cli installer script",
+)
+@click.option(
+    "--oapi", flag_value=True, help="Indicates that the observatory api was installed through the cli installer script"
+)
+@click.option("--editable", flag_value=True, help="Indicates the observatory platform is editable")
+def config(command: str, config_path: str, interactive: bool, ao_wf: bool, oaebu_wf: bool, oapi: bool, editable: bool):
     """Generate config files for the Observatory Platform.\n
 
     COMMAND: the type of config file to generate:\n
       - local: generate a config file for running the Observatory Platform locally.\n
       - terraform: generate a config file for running the Observatory Platform with Terraform.\n
+
+    :param interactive: whether to interactively ask for configuration options.
+    :param ao_wf: Whether academic_observatory_workflows was installed using the installer script.
+    :param oaebu_wf: Whether oaebu_workflows was installed using the installer script.
+    :param oapi: Whether the Observatory API was installed using the installer script.
+    :param editable: Whether the observatory platform is editable.
     """
 
     # Make the generate command, which encapsulates functionality for generating data
     cmd = GenerateCommand()
 
+    if config_path is None:
+        config_path = LOCAL_CONFIG_PATH if command == "local" else TERRAFORM_CONFIG_PATH
+
+    config_name = "Observatory config" if command == "local" else "Terraform config"
+
+    workflows = []
+    if ao_wf:
+        workflows.append("academic-observatory-workflows")
+    if oaebu_wf:
+        workflows.append("oaebu-workflows")
+
     cmd_func = None
-    config_name = ""
-    if command == "local":
-        if config_path is None:
-            config_path = LOCAL_CONFIG_PATH
+    if command == "local" and not interactive:
         cmd_func = cmd.generate_local_config
-        config_name = "Observatory Config"
-    elif command == "terraform":
-        if config_path is None:
-            config_path = TERRAFORM_CONFIG_PATH
+    elif command == "terraform" and not interactive:
         cmd_func = cmd.generate_terraform_config
-        config_name = "Terraform Config"
+    elif command == "local" and interactive:
+        cmd_func = cmd.generate_local_config_interactive
+    else:
+        cmd_func = cmd.generate_terraform_config_interactive
 
     if not os.path.exists(config_path) or click.confirm(
         f'The file "{config_path}" exists, do you want to overwrite it?'
     ):
-        cmd_func(config_path)
+        cmd_func(config_path, workflows=workflows, oapi=oapi, editable=editable)
     else:
         click.echo(f"Not generating {config_name}")
 
@@ -420,7 +548,7 @@ def terraform_check_dependencies(
                 print(indent(f"- {key}: {value}", INDENT3))
     else:
         print(indent("- file not found, generating a default file", INDENT2))
-        generate_cmd.generate_terraform_config(terraform_cmd.config_path)
+        generate_cmd.generate_terraform_config(terraform_cmd.config_path, editable=False, workflows=[], oapi=False)
 
     print(indent("Terraform credentials file:", INDENT1))
     if terraform_cmd.terraform_credentials_exists:
