@@ -23,29 +23,140 @@ import os
 import re
 import time
 from concurrent.futures import ProcessPoolExecutor, as_completed
+from copy import deepcopy
 from enum import Enum
 from multiprocessing import BoundedSemaphore, cpu_count
-from typing import List, Tuple, Union
+from typing import List, Optional, Tuple, Union, Any
 
 import pendulum
+from airflow.models import Variable
 from google.api_core.exceptions import BadRequest, Conflict
 from google.cloud import bigquery, storage
 from google.cloud.bigquery import LoadJob, LoadJobConfig, QueryJob, SourceFormat
+from google.cloud.bigquery.job import QueryJobConfig
 from google.cloud.exceptions import Conflict, NotFound
 from google.cloud.storage import Blob
 from googleapiclient import discovery as gcp_api
-from requests.exceptions import ChunkedEncodingError
-
+from observatory.api.client.model.big_query_bytes_processed import (
+    BigQueryBytesProcessed,
+)
+from observatory.platform.utils.airflow_utils import AirflowVars
+from observatory.platform.utils.api import make_observatory_api
 from observatory.platform.utils.config_utils import utils_templates_path
 from observatory.platform.utils.file_utils import crc32c_base64_hash
 from observatory.platform.utils.jinja2_utils import (
     make_sql_jinja2_filename,
     render_template,
 )
+from requests.exceptions import ChunkedEncodingError
 
 # The chunk size to use when uploading / downloading a blob in multiple parts, must be a multiple of 256 KB.
 DEFAULT_CHUNK_SIZE = 256 * 1024 * 4
 
+# BigQuery daily query byte limit.
+BIGQUERY_QUERY_DAILY_BYTE_LIMIT = 1024 * 1024 * 1024 * 1024 * 5  # 5 TiB
+
+
+# This function makes it easier to mock out in tests. This mechanism is here to allow you to hard code disabling the
+# daily quota check across the platform.  When you are convinced the feature is working correctly in production and
+# that you wan't it always on, then the mechanism can be removed.
+def bq_query_daily_limit_enabled():
+    """Whether to turn on BigQuery daily limits.
+
+    :return: Whether to use daily limits.
+    """
+
+    return True
+
+
+def bq_query_bytes_estimate(*args, **kwargs) -> int:
+    """Do a dry run of a BigQuery query to estimate the bytes processed.
+
+    :param args: Positional arguments to pass onto the callable.
+    :param kwargs: Named arguments to pass onto the callable.
+    :return: Query bytes estimate.
+    """
+
+    if "job_config" not in kwargs:
+        kwargs["job_config"] = QueryJobConfig()
+
+    config = deepcopy(kwargs["job_config"])
+    config.dry_run = True
+    kwargs["job_config"] = config
+
+    bytes_estimate = bigquery.Client().query(*args, **kwargs).total_bytes_processed
+    return bytes_estimate
+
+
+def bq_query_bytes_budget_check(*, bytes_budget: Optional[int], bytes_estimate: int):
+    """Check that the estimated number of processed bytes required does not exceed the budgeted number of bytes for the
+    query. If the estimate exceeds the budget, this function throws an exception.
+
+    :param bytes_budget: The processed bytes budget for this query.
+    :param bytes_estimate: Estimated number of bytes processed in query.
+    """
+
+    if bytes_budget is None:
+        return
+
+    if bytes_estimate > bytes_budget:
+        raise Exception(f"Bytes estimate {bytes_estimate} exceeds the budget {bytes_budget}.")
+
+
+def get_bytes_processed(*, api, project: str) -> int:
+    """Get the bytes processed over the last 24 hours.
+    :param api: Observatory platform API client object.
+    :param project: GCP project.
+    :return: Number of bytes processed.
+    """
+
+    return api.get_bigquery_bytes_processed(project=project)
+
+
+def save_bytes_processed(*, api: Any, project: str, bytes_estimate: int):
+    """Update the number of bytes processed.
+    :param api: Observatory platform API client object.
+    :param project: GCP project.
+    :param bytes_estimate: Bytes estimate of the current query.
+    """
+
+    record = BigQueryBytesProcessed(
+        project=project,
+        total=bytes_estimate,
+    )
+    api.post_bigquery_bytes_processed(record)
+
+
+def bq_query_bytes_daily_limit_check(bytes_estimate: int) -> BigQueryBytesProcessed:
+    """Check the daily byte limit. Raise exception if the current query will put us over the limit.
+
+    :param bytes_estimate: Estimated number of bytes processed in query.
+    """
+
+    # Disable if other repos not ready yet
+    if not bq_query_daily_limit_enabled():
+        return
+
+    try:
+        project = Variable.get(AirflowVars.PROJECT_ID)
+        api = make_observatory_api()
+    except Exception as e:
+        logging.warning(f"Skipping daily byte limit check: {e}")
+        return
+
+    # Get bytes processed over last 24 hours
+    bytes_processed = get_bytes_processed(api=api, project=project)
+
+    # Check that the bytes processed over the last 24 hours + the estimated query amount is below the daily limit
+    projected_estimate = bytes_processed + bytes_estimate
+    if projected_estimate > BIGQUERY_QUERY_DAILY_BYTE_LIMIT:
+        raise Exception(
+            f"The projected bytes estimate of {projected_estimate} exceeds the daily limit of {BIGQUERY_QUERY_DAILY_BYTE_LIMIT}"
+        )
+
+    # Save the estimate, assumes the actual query will run after
+    save_bytes_processed(api=api, project=project, bytes_estimate=bytes_estimate)
+
 
 def table_name_from_blob(blob_name: str, file_extension: str):
     """Make a BigQuery table name from a blob name.
@@ -153,7 +264,7 @@ def load_bigquery_table(
     location: str,
     table: str,
     schema_file_path: str,
-    source_format: SourceFormat,
+    source_format: str,
     csv_field_delimiter: str = ",",
     csv_quote_character: str = '"',
     csv_allow_quoted_newlines: bool = False,
@@ -167,6 +278,7 @@ def load_bigquery_table(
     project_id: str = None,
     cluster: bool = False,
     clustering_fields=None,
+    ignore_unknown_values: bool = False,
 ) -> bool:
     """Load a BigQuery table from an object on Google Cloud Storage.
 
@@ -190,6 +302,7 @@ def load_bigquery_table(
     :param cluster: whether to cluster the table or not.
     :param clustering_fields: what fields to cluster on.
     Default is to overwrite.
+    :param ignore_unknown_values: whether to ignore unknown values or not.
     :return:
     """
 
@@ -219,6 +332,7 @@ def load_bigquery_table(
     job_config.schema = client.schema_from_json(schema_file_path)
     job_config.write_disposition = write_disposition
     job_config.destination_table_description = table_description
+    job_config.ignore_unknown_values = ignore_unknown_values
 
     # Set CSV options
     if source_format == SourceFormat.CSV:
@@ -255,16 +369,25 @@ def load_bigquery_table(
     return state
 
 
-def run_bigquery_query(query: str) -> List:
-    """Run a BigQuery query.
+def run_bigquery_query(query: str, bytes_budget: Optional[int] = 549755813888) -> list:
+    """Run a BigQuery query.  Defaults to 0.5 TiB query budget.
 
     :param query: the query to run.
+    :param bytes_budget: Maximum bytes allowed to be processed by the query.
     :return: the results.
     """
 
+    bytes_estimate = bq_query_bytes_estimate(query)
+
+    bq_query_bytes_budget_check(bytes_budget=bytes_budget, bytes_estimate=bytes_estimate)
+
+    bq_query_bytes_daily_limit_check(bytes_estimate)
+
     client = bigquery.Client()
     query_job = client.query(query)
     rows = query_job.result()
+    success = query_job.errors is None
+
     return list(rows)
 
 
@@ -329,8 +452,9 @@ def create_bigquery_table_from_query(
     require_partition_filter=True,
     cluster: bool = False,
     clustering_fields=None,
+    bytes_budget: Optional[int] = 549755813888,
 ) -> bool:
-    """Create a BigQuery dataset from a provided query.
+    """Create a BigQuery dataset from a provided query. Defaults to 0.5 TiB query budget.
 
     :param sql: the sql query to be executed
     :param labels: labels to place on the new table
@@ -347,6 +471,7 @@ def create_bigquery_table_from_query(
     :param require_partition_filter: whether the partition filter is required or not when querying the table.
     :param cluster: whether to cluster the table or not.
     :param clustering_fields: what fields to cluster on.
+    :param bytes_budget: Maximum bytes allowed to be processed by query.
     :return:
     """
 
@@ -386,14 +511,18 @@ def create_bigquery_table_from_query(
     # Set partitioning settings
     if partition:
         job_config.time_partitioning = bigquery.TimePartitioning(
-            type_=partition_type,
-            field=partition_field,
-            require_partition_filter=require_partition_filter,
+            type_=partition_type, field=partition_field, require_partition_filter=require_partition_filter
         )
 
     if cluster:
         job_config.clustering_fields = clustering_fields
 
+    bytes_estimate = bq_query_bytes_estimate(sql, job_config=job_config)
+
+    bq_query_bytes_budget_check(bytes_budget=bytes_budget, bytes_estimate=bytes_estimate)
+
+    bq_query_bytes_daily_limit_check(bytes_estimate)
+
     query_job: QueryJob = client.query(sql, job_config=job_config)
     query_job.result()
     success = query_job.done()
@@ -697,6 +826,7 @@ def upload_file_to_cloud_storage(
     connection_sem: BoundedSemaphore = None,
     chunk_size: int = DEFAULT_CHUNK_SIZE,
     project_id: str = None,
+    check_blob_hash: bool = True
 ) -> Tuple[bool, bool]:
     """Upload a file to Google Cloud Storage.
 
@@ -707,6 +837,7 @@ def upload_file_to_cloud_storage(
     :param connection_sem: a BoundedSemaphore to limit the number of upload connections that can run at once.
     :param chunk_size: the chunk size to use when uploading a blob in multiple parts, must be a multiple of 256 KB.
     :param project_id: the project in which the bucket is located, defaults to inferred from the environment.
+    :param check_blob_hash: check whether the blob exists and if the crc32c hashes match, in which case skip uploading.
     :return: whether the task was successful or not and whether the file was uploaded.
     """
     func_name = upload_file_to_cloud_storage.__name__
@@ -722,7 +853,7 @@ def upload_file_to_cloud_storage(
     blob = bucket.blob(blob_name)
 
     # Check if blob exists already and matches the file we are uploading
-    if blob.exists():
+    if check_blob_hash and blob.exists():
         # Get blob hash
         blob.reload()
         expected_hash = blob.crc32c
@@ -784,7 +915,7 @@ def google_cloud_storage_transfer_job(job: dict, func_name: str, gc_project_id:
         if "operations" in response:
             operations = response["operations"]
 
-            in_progress_count, success_count, failed_count, aborted_count, objects_count = 0, 0, 0, 0, 0
+            in_progress_count, success_count, failed_count, aborted_count, objects_count = (0, 0, 0, 0, 0)
             for op in operations:
                 status = op["metadata"]["status"]
                 if status == TransferStatus.success.value:
@@ -825,6 +956,7 @@ def google_cloud_storage_transfer_job(job: dict, func_name: str, gc_project_id:
 
 
 def azure_to_google_cloud_storage_transfer(
+    *,
     azure_storage_account_name: str,
     azure_sas_token: str,
     azure_container: str,
@@ -832,6 +964,7 @@ def azure_to_google_cloud_storage_transfer(
     gc_project_id: str,
     gc_bucket: str,
     description: str,
+    gc_bucket_path: str = None,
     start_date: pendulum.DateTime = pendulum.now("UTC"),
 ) -> bool:
     """Transfer files from an Azure blob container to a Google Cloud Storage bucket.
@@ -843,6 +976,7 @@ def azure_to_google_cloud_storage_transfer(
     :param gc_project_id: the Google Cloud project id that holds the Google Cloud Storage bucket.
     :param gc_bucket: the Google Cloud bucket name.
     :param description: a description for the transfer job.
+    :param gc_bucket_path: the path in the Google Cloud bucket to save the objects.
     :param start_date: the date that the transfer job will start.
     :return: whether the transfer was a success or not.
     """
@@ -862,9 +996,7 @@ def azure_to_google_cloud_storage_transfer(
         "transferSpec": {
             "azureBlobStorageDataSource": {
                 "storageAccount": azure_storage_account_name,
-                "azureCredentials": {
-                    "sasToken": azure_sas_token,
-                },
+                "azureCredentials": {"sasToken": azure_sas_token},
                 "container": azure_container,
             },
             "objectConditions": {"includePrefixes": include_prefixes},
@@ -872,6 +1004,13 @@ def azure_to_google_cloud_storage_transfer(
         },
     }
 
+    if gc_bucket_path is not None:
+        # Must end in a / see https://cloud.google.com/storage-transfer/docs/create-manage-transfer-program
+        if not gc_bucket_path.endswith("/"):
+            gc_bucket_path = f"{gc_bucket_path}/"
+
+        job["transferSpec"]["gcsDataSink"]["path"] = gc_bucket_path
+
     success, objects_count = google_cloud_storage_transfer_job(job, func_name, gc_project_id)
     return success
 
@@ -884,8 +1023,10 @@ def aws_to_google_cloud_storage_transfer(
     gc_project_id: str,
     gc_bucket: str,
     description: str,
+    gc_bucket_path: str = None,
     last_modified_since: pendulum.DateTime = None,
     last_modified_before: pendulum.DateTime = None,
+    transfer_manifest: str = None,
     start_date: pendulum.DateTime = pendulum.now("UTC"),
 ) -> Tuple[bool, int]:
     """Transfer files from an AWS bucket to a Google Cloud Storage bucket.
@@ -895,10 +1036,12 @@ def aws_to_google_cloud_storage_transfer(
     :param aws_bucket: the name of the aws S3 bucket where files will be copied from.
     :param include_prefixes: the prefixes of blobs to download from the Azure blob container.
     :param gc_project_id: the Google Cloud project id that holds the Google Cloud Storage bucket.
-    :param last_modified_since:
-    :param last_modified_before:
     :param gc_bucket: the Google Cloud bucket name.
     :param description: a description for the transfer job.
+    :param gc_bucket_path: the destination folder inside the Google Cloud bucket.
+    :param last_modified_since:
+    :param last_modified_before:
+    :param transfer_manifest: Path to manifest file in Google Cloud bucket (incl gs://)
     :param start_date: the date that the transfer job will start.
     :return: whether the transfer was a success or not.
     """
@@ -918,15 +1061,15 @@ def aws_to_google_cloud_storage_transfer(
         "transferSpec": {
             "awsS3DataSource": {
                 "bucketName": aws_bucket,
-                "awsAccessKey": {
-                    "accessKeyId": aws_access_key_id,
-                    "secretAccessKey": aws_secret_key,
-                },
+                "awsAccessKey": {"accessKeyId": aws_access_key_id, "secretAccessKey": aws_secret_key},
             },
             "objectConditions": {"includePrefixes": include_prefixes},
             "gcsDataSink": {"bucketName": gc_bucket},
         },
     }
+    if gc_bucket_path:
+        job["transferSpec"]["gcsDataSink"]["path"] = gc_bucket_path
+
     if last_modified_since:
         job["transferSpec"]["objectConditions"]["lastModifiedSince"] = last_modified_since.isoformat().replace(
             "+00:00", "Z"
@@ -936,6 +1079,9 @@ def aws_to_google_cloud_storage_transfer(
             "+00:00", "Z"
         )
 
+    if transfer_manifest:
+        job["transferSpec"]["transferManifest"] = {"location": transfer_manifest}
+
     success, objects_count = google_cloud_storage_transfer_job(job, func_name, gc_project_id)
     return success, objects_count
 
