@@ -70,14 +70,14 @@ import time
 import unittest
 import uuid
 from dataclasses import dataclass
-from datetime import datetime
-from functools import partial
+from datetime import datetime, timedelta
 from http.server import SimpleHTTPRequestHandler, ThreadingHTTPServer
 from multiprocessing import Process
 from typing import Dict, List
 from unittest.mock import patch
 
 import croniter
+import google
 import httpretty
 import paramiko
 import pendulum
@@ -96,6 +96,7 @@ from click.testing import CliRunner
 from cryptography.hazmat.backends import default_backend
 from cryptography.hazmat.primitives import serialization
 from cryptography.hazmat.primitives.asymmetric import rsa
+from dateutil.relativedelta import relativedelta
 from freezegun import freeze_time
 from google.cloud import bigquery, storage
 from google.cloud.exceptions import NotFound
@@ -104,8 +105,8 @@ from observatory.platform.elastic.elastic_environment import ElasticEnvironment
 from observatory.platform.utils.airflow_utils import AirflowVars
 from observatory.platform.utils.config_utils import module_file_path
 from observatory.platform.utils.file_utils import (
-    _hash_file,
     crc32c_base64_hash,
+    get_file_hash,
     gzip_file_crc,
     list_to_jsonl_gz,
 )
@@ -289,6 +290,10 @@ class ObservatoryEnvironment:
             bucket.delete(force=True)
         except requests.exceptions.ReadTimeout:
             pass
+        except google.api_core.exceptions.NotFound:
+            logging.warning(
+                f"Bucket {bucket_id} not found. Did you mean to call _delete_bucket on the same bucket twice?"
+            )
 
     def add_dataset(self, prefix: str = "") -> str:
         """Add a BigQuery dataset to the Observatory environment.
@@ -340,27 +345,23 @@ class ObservatoryEnvironment:
         self.session.add(conn)
         self.session.commit()
 
-    def run_task(self, task_id: str, dag: DAG = None, execution_date: pendulum.DateTime = None) -> TaskInstance:
+    def run_task(self, task_id: str) -> TaskInstance:
         """Run an Airflow task.
 
-        :param dag: the Airflow DAG instance.
         :param task_id: the Airflow task identifier.
-        :param execution_date: the execution date of the DAG.
         :return: None.
         """
-        # If dag or execution date are not set, get corresponding values from dag run
-        if not (dag and execution_date):
-            if self.dag_run is None:
-                raise TypeError(
-                    "Either dag and execution date should be set, or task should be run within a DagRun environment."
-                )
-            dag = self.dag_run.dag
-            execution_date = self.dag_run.execution_date
+
+        assert self.dag_run is not None, "with create_dag_run must be called before run_task"
+
+        dag = self.dag_run.dag
+        run_id = self.dag_run.run_id
         task = dag.get_task(task_id=task_id)
-        ti = TaskInstance(task, execution_date)
-        ti.refresh_from_db()
+        ti = TaskInstance(task, run_id=run_id)
+        ti.dag_run = self.dag_run
         ti.init_run_context(raw=True)
         ti.run(ignore_ti_state=True)
+
         return ti
 
     @contextlib.contextmanager
@@ -374,7 +375,12 @@ class ObservatoryEnvironment:
         :return: None.
         """
         # Get start date, which is one schedule interval after execution date
-        start_date = croniter.croniter(dag.normalized_schedule_interval, execution_date).get_next(pendulum.DateTime)
+        if isinstance(dag.normalized_schedule_interval, (timedelta, relativedelta)):
+            start_date = (
+                datetime.fromtimestamp(execution_date.timestamp(), pendulum.tz.UTC) + dag.normalized_schedule_interval
+            )
+        else:
+            start_date = croniter.croniter(dag.normalized_schedule_interval, execution_date).get_next(pendulum.DateTime)
         frozen_time = freeze_time(start_date, tick=True)
 
         run_id = "manual__{0}".format(execution_date.isoformat())
@@ -388,7 +394,7 @@ class ObservatoryEnvironment:
                 self.dag_run = dag.create_dagrun(
                     run_id=run_id, state=state, execution_date=execution_date, start_date=pendulum.now("UTC")
                 )
-                yield
+                yield self.dag_run
             finally:
                 self.dag_run.update_state()
                 if freeze:
@@ -545,12 +551,25 @@ class ObservatoryTestCase(unittest.TestCase):
         """
 
         with CliRunner().isolated_filesystem() as dag_folder:
-            if os.path.exists(dag_file):
-                shutil.copy(dag_file, os.path.join(dag_folder, os.path.basename(dag_file)))
+            if not os.path.exists(dag_file):
+                raise Exception(f"{dag_file} does not exist.")
+
+            shutil.copy(dag_file, os.path.join(dag_folder, os.path.basename(dag_file)))
+
             dag_bag = DagBag(dag_folder=dag_folder)
-            dag = dag_bag.get_dag(dag_id=dag_id)
+
+            if dag_bag.import_errors != {}:
+                logging.error(f"DagBag errors: {dag_bag.import_errors}")
             self.assertEqual({}, dag_bag.import_errors)
+
+            dag = dag_bag.get_dag(dag_id=dag_id)
+
+            if dag is None:
+                logging.error(
+                    f"DAG not found in the database. Make sure the DAG ID is correct, and the dag file contains the words 'airflow' and 'DAG'."
+                )
             self.assertIsNotNone(dag)
+
             self.assertGreaterEqual(len(dag.tasks), 1)
 
     def assert_blob_exists(self, bucket_id: str, blob_name: str):
@@ -612,6 +631,51 @@ class ObservatoryTestCase(unittest.TestCase):
         if expected_rows is not None:
             self.assertEqual(expected_rows, actual_rows)
 
+    def assert_table_content(self, table_id: str, expected_content: List[dict] = None):
+        """Assert whether a BigQuery table has any content and if expected content is given whether it matches the
+        actual content. The order of the rows is not checked, only whether all rows in the expected content match
+        the rows in the actual content.
+        The expected content should be a list of dictionaries, where each dictionary represents one row of the table,
+        the keys are fieldnames and values are values.
+
+        :param table_id: the BigQuery table id.
+        :param expected_content: the expected content.
+        :return: whether the table has content and the expected content is correct
+        """
+        rows = None
+        actual_content = None
+        try:
+            rows = self.bigquery_client.list_rows(table_id)
+            actual_content = [dict(row) for row in rows]
+        except NotFound:
+            pass
+
+        self.assertIsNotNone(rows)
+        if expected_content is not None:
+            for row in expected_content:
+                self.assertIn(row, actual_content)
+                actual_content.remove(row)
+            self.assertListEqual(
+                [], actual_content, msg=f"Rows in actual content that are not in expected content: {actual_content}"
+            )
+
+    def assert_table_bytes(self, table_id: str, expected_bytes: int):
+        """Assert whether the given bytes from a BigQuery table matches the expected bytes.
+
+        :param table_id: the BigQuery table id.
+        :param expected_bytes: the expected number of bytes.
+        :return: whether the table exists and the expected bytes match
+        """
+
+        table = None
+        try:
+            table = self.bigquery_client.get_table(table_id)
+        except NotFound:
+            pass
+
+        self.assertIsNotNone(table)
+        self.assertEqual(expected_bytes, table.num_bytes)
+
     def assert_file_integrity(self, file_path: str, expected_hash: str, algorithm: str):
         """Assert that a file exists and it has the correct hash.
 
@@ -623,14 +687,11 @@ class ObservatoryTestCase(unittest.TestCase):
 
         self.assertTrue(os.path.isfile(file_path))
 
-        if algorithm == "md5":
-            hash_func = partial(_hash_file, algorithm="md5")
-        elif algorithm == "gzip_crc":
-            hash_func = gzip_file_crc
+        if algorithm == "gzip_crc":
+            actual_hash = gzip_file_crc(file_path)
         else:
-            raise ValueError(f"Unknown hash algorithm: {algorithm}")
+            actual_hash = get_file_hash(file_path=file_path, algorithm=algorithm)
 
-        actual_hash = hash_func(file_path)
         self.assertEqual(expected_hash, actual_hash)
 
     def assert_cleanup(self, download_folder: str, extract_folder: str, transform_folder: str):
@@ -848,7 +909,7 @@ def bq_load_tables(
 
         # Save to JSONL
         for table in tables:
-            blob_name = f"{table.table_name}.jsonl.gz"
+            blob_name = f"{table.dataset_id}-{table.table_name}.jsonl.gz"
             file_path = os.path.join(t, blob_name)
             list_to_jsonl_gz(file_path, table.records)
             files_list.append(file_path)
