@@ -22,38 +22,31 @@ import logging
 import os
 import pathlib
 import re
-import sys
 import traceback
 from base64 import b64decode
-from collections import deque
-from dataclasses import dataclass
 from datetime import datetime, timedelta
 from enum import Enum
-from math import ceil
-from pathlib import Path
-from typing import Any, List, Optional, Tuple, Type, Union
+from typing import Any, List, Optional, Tuple, Union
 
-import jsonlines
 import paramiko
 import pendulum
 import pysftp
 import six
 from airflow import AirflowException
 from airflow.hooks.base import BaseHook
-from airflow.models import DagBag, Variable
-from airflow.models.taskinstance import TaskInstance
+from airflow.models import DagBag, DagRun, Variable, XCom
 from airflow.secrets.environment_variables import EnvironmentVariablesBackend
+from airflow.utils.db import provide_session
 from dateutil.relativedelta import relativedelta
 from google.cloud import bigquery
-from google.cloud.bigquery import SourceFormat
 from observatory.platform.observatory_config import Environment
 from observatory.platform.utils.airflow_utils import (
     AirflowConns,
     AirflowVars,
-    create_slack_webhook,
+    send_slack_msg,
 )
+from observatory.platform.utils.api import make_observatory_api
 from observatory.platform.utils.config_utils import find_schema, utils_templates_path
-from observatory.platform.utils.file_utils import load_file, write_to_file
 from observatory.platform.utils.gc_utils import (
     bigquery_sharded_table_id,
     copy_bigquery_table,
@@ -61,13 +54,14 @@ from observatory.platform.utils.gc_utils import (
     load_bigquery_table,
     run_bigquery_query,
     select_table_shard_dates,
-    upload_file_to_cloud_storage,
     upload_files_to_cloud_storage,
 )
 from observatory.platform.utils.jinja2_utils import (
     make_sql_jinja2_filename,
     render_template,
 )
+from sqlalchemy import and_
+from sqlalchemy.orm import Session
 
 ScheduleInterval = Union[str, timedelta, relativedelta]
 
@@ -301,7 +295,7 @@ def bq_load_shard(
     transform_blob: str,
     dataset_id: str,
     table_id: str,
-    source_format: SourceFormat,
+    source_format: str,
     prefix: str = "",
     schema_version: str = None,
     dataset_description: str = "",
@@ -346,7 +340,7 @@ def bq_load_shard_v2(
     dataset_location: str,
     table_id: str,
     release_date: pendulum.Date,
-    source_format: SourceFormat,
+    source_format: str,
     prefix: str = "",
     schema_version: str = None,
     dataset_description: str = "",
@@ -409,7 +403,7 @@ def bq_load_ingestion_partition(
     dataset_id: str,
     main_table_id: str,
     partition_table_id: str,
-    source_format: SourceFormat,
+    source_format: str,
     prefix: str = "",
     schema_version: str = None,
     dataset_description: str = "",
@@ -420,7 +414,7 @@ def bq_load_ingestion_partition(
     given it will automatically partition by ingestion datetime.
 
     :param schema_folder: the path to the SQL schema folder.
-    :param end_date: End date.
+    :param end_date: Release end date, used to find the schema and to create table id.
     :param transform_blob: Name of the transform blob.
     :param dataset_id: Dataset id.
     :param main_table_id: Main table id.
@@ -439,7 +433,7 @@ def bq_load_ingestion_partition(
     uri = f"gs://{bucket_name}/{transform_blob}"
 
     # Include date in table id, so data in table is not overwritten
-    partition_table_id = create_date_table_id(partition_table_id, pendulum.today(), partition_type)
+    partition_table_id = create_date_table_id(partition_table_id, end_date, partition_type)
     success = load_bigquery_table(
         uri,
         dataset_id,
@@ -464,7 +458,7 @@ def bq_load_partition(
     dataset_location: str,
     table_id: str,
     release_date: pendulum.DateTime,
-    source_format: SourceFormat,
+    source_format: str,
     partition_type: bigquery.TimePartitioningType,
     prefix: str = "",
     schema_version: str = None,
@@ -525,26 +519,26 @@ def bq_load_partition(
 
 
 def bq_delete_old(
-    start_date: pendulum.DateTime,
-    end_date: pendulum.DateTime,
+    ingestion_date: pendulum.DateTime,
     dataset_id: str,
     main_table_id: str,
     partition_table_id: str,
     merge_partition_field: str,
+    bytes_budget: Optional[int] = None,
 ):
     """Will run a BigQuery query that deletes rows from the main table that are matched with rows from
-    specific partitions of the partition table.
+    a specific partition of the partition table.
     The query is created from a template and the given info.
-    :param start_date: Start date, excluded.
-    :param end_date: End date, included.
+
+    :param ingestion_date: The ingestion date of the partition that will be used.
     :param dataset_id: Dataset id.
     :param main_table_id: Main table id.
     :param partition_table_id: Partition table id.
     :param merge_partition_field: Merge partition field.
+    :param bytes_budget: Maximum bytes allowed to be processed.
     :return: None.
     """
-    start_date = start_date.strftime("%Y-%m-%d")
-    end_date = end_date.strftime("%Y-%m-%d")
+    ingestion_date = ingestion_date.strftime("%Y-%m-%d")
     # Get merge variables
     dataset_id = dataset_id
     main_table = main_table_id
@@ -558,50 +552,36 @@ def bq_delete_old(
         main_table=main_table,
         partitioned_table=partitioned_table,
         merge_condition_field=merge_condition_field,
-        start_date=start_date,
-        end_date=end_date,
+        ingestion_date=ingestion_date,
     )
-    run_bigquery_query(query)
+    run_bigquery_query(query, bytes_budget=bytes_budget)
 
 
 def bq_append_from_partition(
-    schema_folder: str,
-    start_date: pendulum.DateTime,
-    end_date: pendulum.DateTime,
+    ingestion_date: pendulum.DateTime,
     dataset_id: str,
     main_table_id: str,
     partition_table_id: str,
-    prefix: str = "",
-    schema_version: str = None,
 ):
-    """Appends rows to the main table by coping specific partitions from the partition table to the main table.
-    :param schema_folder: the SQL schema path.
-    :param start_date: Start date, excluded.
-    :param end_date: End date, included.
+    """Appends rows to the main table by coping a specific partition from the partition table to the main table.
+
+    :param ingestion_date: The ingestion date of the partition that will be used.
     :param dataset_id: Dataset id.
     :param main_table_id: Main table id.
     :param partition_table_id: Partition table id.
-    :param prefix: The prefix for the schema.
-    :param schema_version: Schema version.
     :return: None.
     """
-    project_id, bucket_name, data_location, schema_file_path = prepare_bq_load(
-        schema_folder, dataset_id, main_table_id, end_date, prefix, schema_version
-    )
-    # exclude start date and include end date in period
-    period = pendulum.period(start_date.date() + timedelta(days=1), end_date.date())
-    logging.info(f"Getting table partitions: ")
-    source_table_ids = []
-    for dt in period:
-        table_id = f"{project_id}.{dataset_id}.{partition_table_id}${dt.strftime('%Y%m%d')}"
-        source_table_ids.append(table_id)
-        logging.info(f"Adding table_id: {table_id}")
-    destination_table_id = f"{project_id}.{dataset_id}.{main_table_id}"
-    success = copy_bigquery_table(
-        source_table_ids, destination_table_id, data_location, bigquery.WriteDisposition.WRITE_APPEND
-    )
+    logging.info("requesting project_id variable")
+    project_id = Variable.get(AirflowVars.PROJECT_ID)
+
+    logging.info("requesting data_location variable")
+    data_location = Variable.get(AirflowVars.DATA_LOCATION)
+
+    src_table_id = f"{project_id}.{dataset_id}.{partition_table_id}${ingestion_date.strftime('%Y%m%d')}"
+    dst_table_id = f"{project_id}.{dataset_id}.{main_table_id}"
+    success = copy_bigquery_table(src_table_id, dst_table_id, data_location, bigquery.WriteDisposition.WRITE_APPEND)
     if not success:
-        raise AirflowException()
+        raise AirflowException("Error copying BigQuery table")
 
 
 def bq_append_from_file(
@@ -610,7 +590,7 @@ def bq_append_from_file(
     transform_blob: str,
     dataset_id: str,
     main_table_id: str,
-    source_format: SourceFormat,
+    source_format: str,
     prefix: str = "",
     schema_version: str = None,
     dataset_description: str = "",
@@ -652,10 +632,13 @@ def bq_append_from_file(
         raise AirflowException()
 
 
-def on_failure_callback(**kwargs):
+def on_failure_callback(context):
     """
     Function that is called on failure of an airflow task. Will create a slack webhook and send a notification.
-    :param kwargs:
+
+    :param context: the context passed from the PythonOperator. See
+    https://airflow.apache.org/docs/stable/macros-ref.html for a list of the keyword arguments that are passed to
+    this  argument.
     :return: None.
     """
 
@@ -668,13 +651,18 @@ def on_failure_callback(**kwargs):
     if environment == Environment.develop.value:
         logging.info("Not sending slack notification in develop environment.")
     else:
-        exception = kwargs.get("exception")
-        formatted_exception = "".join(
-            traceback.format_exception(etype=type(exception), value=exception, tb=exception.__traceback__)
-        ).strip()
+        exception = context.get("exception")
+        if isinstance(exception, Exception):
+            formatted_exception = "".join(
+                traceback.format_exception(etype=type(exception), value=exception, tb=exception.__traceback__)
+            ).strip()
+        else:
+            formatted_exception = exception
+
         comments = f"Task failed, exception:\n{formatted_exception}"
-        slack_hook = create_slack_webhook(comments, project_id, **kwargs)
-        slack_hook.execute()
+        ti = context["ti"]
+        execution_date = context["execution_date"]
+        send_slack_msg(ti=ti, execution_date=execution_date, comments=comments, project_id=project_id)
 
 
 class SubFolder(Enum):
@@ -838,45 +826,6 @@ class SftpFolders:
                 sftp.rename(in_progress_file, finished_file)
 
 
-def make_observatory_api() -> "ObservatoryApi":  # noqa: F821
-    """Make the ObservatoryApi object, configuring it with a host and api_key.
-
-    :return: the ObservatoryApi.
-    """
-
-    try:
-        from observatory.api.client.api.observatory_api import ObservatoryApi
-        from observatory.api.client.api_client import ApiClient
-        from observatory.api.client.configuration import Configuration
-    except ImportError as e:
-        logging.error("Please install the observatory-api Python package to use the make_observatory_api function")
-        raise e
-
-    # Get connection
-    conn = BaseHook.get_connection(AirflowConns.OBSERVATORY_API)
-
-    # Assert connection has required fields
-    assert (
-        conn.conn_type != "" and conn.conn_type is not None
-    ), f"Airflow Connection {AirflowConns.OBSERVATORY_API} conn_type must not be None"
-    assert (
-        conn.host != "" and conn.host is not None
-    ), f"Airflow Connection {AirflowConns.OBSERVATORY_API} host must not be None"
-    assert (
-        conn.password != "" and conn.password is not None
-    ), f"Airflow Connection {AirflowConns.OBSERVATORY_API} password must not be None"
-
-    # Make host
-    host = f'{str(conn.conn_type).replace("_", "-").lower()}://{conn.host}'
-    if conn.port:
-        host += f":{conn.port}"
-
-    # Return ObservatoryApi
-    config = Configuration(host=host, api_key={"api_key": conn.password})
-    api_client = ApiClient(config)
-    return ObservatoryApi(api_client=api_client)
-
-
 def make_dag_id(namespace: str, organisation_name: str) -> str:
     """Make a DAG id from a namespace and an organisation name.
 
@@ -959,30 +908,6 @@ def build_schedule(sched_start_date: pendulum.DateTime, sched_end_date: pendulum
     return schedule
 
 
-def delete_msg_files(ti: TaskInstance, topic: str, task_id: str, msg_key: str = ""):
-    """Pull messages from a topic and delete the relevant paths.
-
-    :param ti: TaskInstance.
-    :param topic: Message topic.
-    :param task_id: Task ID who sent message.
-    :param msg_key: Key of specific messages.
-    """
-
-    msgs = ti.xcom_pull(key=topic, task_ids=task_id, include_prior_dates=True)
-
-    if msg_key != "":
-        files = msgs[msg_key]
-    else:
-        files = msgs
-
-    for file in files:
-        try:
-            logging.info(f"delete_msg_files: Deleting {file}")
-            Path(file).unlink()
-        except FileNotFoundError as e:
-            logging.warning(f"No such file or directory {file}: {e}")
-
-
 def get_as_list(base: dict, target):
     """Helper function that returns the target as a list.
 
@@ -1033,222 +958,6 @@ def get_entry_or_none(base: dict, target, var_type=None):
     return base[target]
 
 
-def json_to_db(
-    json_list: List[Tuple[Any]], release_date: str, parser, institutes: List[str], path_prefix: str = None
-) -> List[str]:
-    """Transform json from query into database format.
-
-    :param json_list: json data to transform.
-    :param release_date: release date of the snapshot.
-    :param parser: Parser function accepting (json entry, harvest date, release date, institutes), and returning
-                    schema conforming data structure.
-    :param institutes: List of institution ids used in the query.
-    :param path_prefix: If specified, gives the new path prefix for the file you want to save.
-    :return: Saved file name.
-    """
-
-    jsonlines_files = list()
-    if len(json_list) == 0:
-        return jsonlines_files
-
-    first_file = json_list[0][0]
-    save_file = f"{first_file}l"
-
-    if path_prefix:
-        filename = os.path.basename(save_file)
-        save_file = os.path.join(path_prefix, filename)
-        Path(path_prefix).mkdir(parents=True, exist_ok=True)
-
-    jsonlines_files.append(save_file)
-
-    with jsonlines.open(save_file, mode="w") as writer:
-        for (file, harvest_date) in json_list:
-            logging.info(f"Parsing {file} into db format and writing to jsonlines")
-            with open(file, "r") as f:
-                data = json.load(f)
-
-            parsed_entries = list()
-            for entry in data:
-                if not isinstance(entry, dict):
-                    continue
-                parsed_entry = parser(entry, harvest_date, release_date, institutes)
-                parsed_entries.append(parsed_entry)
-
-            for entry in parsed_entries:
-                writer.write(entry)
-
-    return jsonlines_files
-
-
-def validate_date(date_string):
-    """Validate a date string is pendulum parsable.
-
-    :param date_string: date string to check.
-    :return: True if parsable, false otherwise.
-    """
-    try:
-        pendulum.parse(date_string)
-    except Exception as e:
-        print(f"Pendulum parsing encountered exception: {e}")
-        return False
-    return True
-
-
-def write_xml_to_json(transform_path: str, release_date: str, inst_id: str, in_files: List[str], parser):
-    """Write a list of web responses to json.
-
-    :param transform_path: base path to store transformed files.
-    :param release_date: release date.
-    :param inst_id: institution id from airflow connection id.
-    :param in_files: list of xml web response files.
-    :param parser: Parsing function that parses the response into json compatible data.
-    :return: List of json files written to, and list of schema versions per response.
-    """
-
-    json_file_list = list()
-    schema_vers = list()
-
-    for file in in_files:
-        logging.info(f"Transforming {file} to json")
-        xml_data = load_file(file)
-
-        parsed_list = list()
-        parsed_record, schema_ver = parser(xml_data)
-        if parsed_record is None:
-            logging.info(f"Empty record received for {file}")
-            continue
-        parsed_list = parsed_list + parsed_record
-        schema_vers.append(schema_ver)
-
-        # Save it in the transform bucket.
-        filename = os.path.basename(file)
-        json_file = f"{filename[:-3]}json"
-        json_path = os.path.join(transform_path, release_date, inst_id, json_file)
-        json_file_list.append(json_path)
-        json_record = json.dumps(parsed_list)
-        write_to_file(json_record, json_path)
-
-    return json_file_list, schema_vers
-
-
-@dataclass
-class PeriodCount:
-    """Descriptive wrapper for a (period, count) object."""
-
-    period: pendulum.Period  # The schedule period in question
-    count: int  # Number of results for this period.
-
-
-class ScheduleOptimiser:
-    """Calculate a schedule that minimises API calls using historic retrieval data.
-    Given a list of tuples (period, count) that indicates how many results were retrieved for a given period from a
-    historical query, the maximum number of results per API call, and the maximum number of results per query, get
-    a schedule that minimises the number of API calls made.
-    """
-
-    @staticmethod
-    def get_num_calls(num_results: int, max_per_call: int) -> int:
-        """Calculate the number of required API calls based on number of results and max results per call.
-
-        :param num_results: Number of results.
-        :param max_per_call: The max returnable results per call.
-        :return: Number of API calls you need to make.
-        """
-
-        return ceil(float(num_results) / max_per_call)
-
-    @staticmethod
-    def extract_schedule(historic_counts: List[PeriodCount], moves: List[int]) -> List[Type[pendulum.Period]]:
-        """Extract a solution schedule from the optimisation.
-
-        :param historic_counts: the histogram of periods and their counts.
-        :param moves: the moves the optimiser took to compute the minimum.
-        :return: Optimised schedule.
-        """
-
-        stack = deque()
-
-        j = len(moves) - 1
-        while j >= 0:
-            i = moves[j]
-            period = pendulum.Period(historic_counts[i].period.start, historic_counts[j].period.end)
-            stack.append(period)
-            j = i - 1
-
-        schedule = list()
-        while len(stack) > 0:
-            schedule.append(stack.pop())
-
-        return schedule
-
-    @staticmethod
-    def optimise(
-        max_per_call: int, max_per_query: int, historic_counts: List[Type[PeriodCount]]
-    ) -> Tuple[List[Type[pendulum.Period]], int]:
-        """Calculate and return a schedule that minimises the number of API calls with the given constraints. Behaviour
-            if there are 0 results in any of the periods is still to return 1 period covering the entire span, but the
-            minimum number of calls reported will be 0.
-
-        :param max_per_call: Maximum number of results returned per API call.
-        :param max_per_query: Maximum number of results returned per query.
-        :param historic_counts: List of results per period, i.e., tuples of form (period, count). Please sort by
-                                date beforehand.
-        :return: New schedule of periods that minimises API calls, and the api calls required for it.
-        """
-
-        n = len(historic_counts)
-
-        if n == 0:
-            raise Exception("Empty historic_counts received.")
-
-        if n == 1:
-            return historic_counts, ScheduleOptimiser.get_num_calls(historic_counts[0].count, max_per_call)
-
-        min_calls = [sys.maxsize] * n
-        moves = [0] * n
-        min_calls[0] = ScheduleOptimiser.get_num_calls(historic_counts[0].count, max_per_call)
-
-        for i in range(1, n):
-            result_count = 0
-            min_calls[i] = ScheduleOptimiser.get_num_calls(historic_counts[i].count, max_per_call) + min_calls[i - 1]
-
-            for j in range(i, -1, -1):
-                curr_count = historic_counts[j].count
-                result_count += curr_count
-                if result_count > max_per_query:
-                    break
-
-                candidate = ScheduleOptimiser.get_num_calls(result_count, max_per_call)
-                if j - 1 >= 0:
-                    candidate += min_calls[j - 1]
-
-                if candidate <= min_calls[i]:
-                    min_calls[i] = candidate
-                    moves[i] = j
-
-        schedule = ScheduleOptimiser.extract_schedule(historic_counts, moves)
-        return schedule, min_calls[-1]
-
-
-def upload_telescope_file_list(bucket_name: str, inst_id: str, telescope_path: str, file_list: List[str]) -> List[str]:
-    """Upload list of files to cloud storage.
-
-    :param bucket_name: Name of storage bucket.
-    :param inst_id: institution id from airflow connection id.
-    :param telescope_path: Path to upload telescope data.
-    :param file_list: List of files to upload.
-    :return: List of location paths in the cloud.
-    """
-
-    blob_list = list()
-    for file in file_list:
-        file_name = os.path.basename(file)
-        blob_name = f"{telescope_path}/{inst_id}/{file_name}"
-        blob_list.append(blob_name)
-        upload_file_to_cloud_storage(bucket_name, blob_name, file_path=file)
-    return blob_list
-
-
 def add_partition_date(
     list_of_dicts: List[dict],
     partition_date: datetime,
@@ -1277,8 +986,17 @@ def add_partition_date(
 def make_release_date(**kwargs) -> pendulum.DateTime:
     """Make a release date"""
 
-    d = kwargs["next_execution_date"].subtract(microseconds=1).date()
-    return pendulum.datetime(d.year, d.month, d.day)
+    return kwargs["next_execution_date"].subtract(days=1).start_of("day")
+
+
+def is_first_dag_run(dag_run: DagRun) -> bool:
+    """Whether the DAG Run is the first run or not
+
+    :param dag_run: A Dag Run instance
+    :return: Whether the DAG run is the first run or not
+    """
+
+    return dag_run.get_previous_dagrun() is None
 
 
 def make_table_name(
@@ -1316,3 +1034,28 @@ def get_chunks(*, input_list: List[Any], chunk_size: int = 8) -> List[Any]:
     n = len(input_list)
     for i in range(0, n, chunk_size):
         yield input_list[i : i + chunk_size]
+
+
+@provide_session
+def delete_old_xcoms(
+    session: Session = None,
+    dag_id: str = None,
+    execution_date: pendulum.DateTime = None,
+    retention_days: int = 31,
+):
+    """Delete XCom messages created by the DAG with the given ID that are as old or older than than
+    execution_date - retention_days.  Defaults to 31 days of retention.
+
+    :param session: DB session.
+    :param dag_id: DAG ID.
+    :param execution_date: DAG execution date.
+    :param retention_days: Days of messages to retain.
+    """
+
+    cut_off_date = execution_date.subtract(days=retention_days)
+    session.query(XCom).filter(
+        and_(
+            XCom.dag_id == dag_id,
+            XCom.execution_date <= cut_off_date,
+        )
+    ).delete()
